@@ -50,8 +50,6 @@ struct connman_ipconfig {
 	enum connman_ipconfig_method method;
 	struct connman_ipaddress *address;
 	struct connman_ipaddress *system;
-
-	struct connman_ipconfig *ipv6;
 };
 
 struct connman_ipdevice {
@@ -74,13 +72,10 @@ struct connman_ipdevice {
 	char *ipv4_gateway;
 	char *ipv6_gateway;
 
-	char *proxy;
 	char *pac;
 
-	struct connman_ipconfig *config;
-
-	struct connman_ipconfig_driver *driver;
-	struct connman_ipconfig *driver_config;
+	struct connman_ipconfig *config_ipv4;
+	struct connman_ipconfig *config_ipv6;
 };
 
 static GHashTable *ipdevice_hash = NULL;
@@ -116,16 +111,23 @@ void connman_ipaddress_free(struct connman_ipaddress *ipaddress)
 	g_free(ipaddress);
 }
 
-static unsigned char netmask2prefixlen(const char *netmask)
+unsigned char __connman_ipconfig_netmask_prefix_len(const char *netmask)
 {
-	unsigned char bits = 0;
-	in_addr_t mask = inet_network(netmask);
-	in_addr_t host = ~mask;
+	unsigned char bits;
+	in_addr_t mask;
+	in_addr_t host;
+
+	if (netmask == NULL)
+		return 32;
+
+	mask = inet_network(netmask);
+	host = ~mask;
 
 	/* a valid netmask must be 2^n - 1 */
 	if ((host & (host + 1)) != 0)
 		return -1;
 
+	bits = 0;
 	for (; mask; mask <<= 1)
 		++bits;
 
@@ -177,10 +179,7 @@ void connman_ipaddress_set_ipv4(struct connman_ipaddress *ipaddress,
 	if (ipaddress == NULL)
 		return;
 
-	if (netmask != NULL)
-		ipaddress->prefixlen = netmask2prefixlen(netmask);
-	else
-		ipaddress->prefixlen = 32;
+	ipaddress->prefixlen = __connman_ipconfig_netmask_prefix_len(netmask);
 
 	g_free(ipaddress->local);
 	ipaddress->local = g_strdup(address);
@@ -299,13 +298,19 @@ static void free_ipdevice(gpointer data)
 	connman_info("%s {remove} index %d", ipdevice->ifname,
 							ipdevice->index);
 
-	if (ipdevice->config != NULL)
-		connman_ipconfig_unref(ipdevice->config);
+	if (ipdevice->config_ipv4 != NULL) {
+		connman_ipconfig_unref(ipdevice->config_ipv4);
+		ipdevice->config_ipv4 = NULL;
+	}
+
+	if (ipdevice->config_ipv6 != NULL) {
+		connman_ipconfig_unref(ipdevice->config_ipv6);
+		ipdevice->config_ipv6 = NULL;
+	}
 
 	free_address_list(ipdevice);
 	g_free(ipdevice->ipv4_gateway);
 	g_free(ipdevice->ipv6_gateway);
-	g_free(ipdevice->proxy);
 	g_free(ipdevice->pac);
 
 	g_free(ipdevice->address);
@@ -313,114 +318,32 @@ static void free_ipdevice(gpointer data)
 	g_free(ipdevice);
 }
 
-static GSList *driver_list = NULL;
-
-static gint compare_priority(gconstpointer a, gconstpointer b)
-{
-	const struct connman_ipconfig_driver *driver1 = a;
-	const struct connman_ipconfig_driver *driver2 = b;
-
-	return driver2->priority - driver1->priority;
-}
-
-/**
- * connman_ipconfig_driver_register:
- * @driver: IP configuration driver
- *
- * Register a new IP configuration driver
- *
- * Returns: %0 on success
- */
-int connman_ipconfig_driver_register(struct connman_ipconfig_driver *driver)
-{
-	DBG("driver %p name %s", driver, driver->name);
-
-	driver_list = g_slist_insert_sorted(driver_list, driver,
-							compare_priority);
-
-	return 0;
-}
-
-/**
- * connman_ipconfig_driver_unregister:
- * @driver: IP configuration driver
- *
- * Remove a previously registered IP configuration driver.
- */
-void connman_ipconfig_driver_unregister(struct connman_ipconfig_driver *driver)
-{
-	DBG("driver %p name %s", driver, driver->name);
-
-	driver_list = g_slist_remove(driver_list, driver);
-}
-
 static void __connman_ipconfig_lower_up(struct connman_ipdevice *ipdevice)
 {
-	GSList *list;
-
-	DBG("ipconfig %p", ipdevice->config);
-
-	if (ipdevice->config == NULL)
-		return;
-
-	switch (ipdevice->config->method) {
-	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
-	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_FIXED:
-	case CONNMAN_IPCONFIG_METHOD_MANUAL:
-		return;
-	case CONNMAN_IPCONFIG_METHOD_DHCP:
-		break;
-	}
-
-	if (ipdevice->driver != NULL)
-		return;
-
-	ipdevice->driver_config = connman_ipconfig_clone(ipdevice->config);
-	if (ipdevice->driver_config == NULL)
-		return;
-
-	for (list = driver_list; list; list = list->next) {
-		struct connman_ipconfig_driver *driver = list->data;
-
-		if (driver->request(ipdevice->driver_config) == 0) {
-			ipdevice->driver = driver;
-			break;
-		}
-	}
-
-	if (ipdevice->driver == NULL) {
-		connman_ipconfig_unref(ipdevice->driver_config);
-		ipdevice->driver_config = NULL;
-	}
+	DBG("ipconfig ipv4 %p ipv6 %p", ipdevice->config_ipv4,
+					ipdevice->config_ipv6);
 }
 
 static void __connman_ipconfig_lower_down(struct connman_ipdevice *ipdevice)
 {
-	DBG("ipconfig %p", ipdevice->config);
+	DBG("ipconfig ipv4 %p ipv6 %p", ipdevice->config_ipv4,
+					ipdevice->config_ipv6);
 
-	if (ipdevice->config == NULL)
-		return;
+	if (ipdevice->config_ipv4)
+		connman_inet_clear_address(ipdevice->index,
+					ipdevice->config_ipv4->address);
 
-	if (ipdevice->driver == NULL)
-		return;
-
-	ipdevice->driver->release(ipdevice->driver_config);
-
-	ipdevice->driver = NULL;
-
-	connman_ipconfig_unref(ipdevice->driver_config);
-	ipdevice->driver_config = NULL;
-
-	connman_inet_clear_address(ipdevice->index);
-	connman_inet_clear_ipv6_address(ipdevice->index,
-			ipdevice->driver_config->address->local,
-			ipdevice->driver_config->address->prefixlen);
+	if (ipdevice->config_ipv6)
+		connman_inet_clear_ipv6_address(ipdevice->index,
+				ipdevice->config_ipv6->address->local,
+				ipdevice->config_ipv6->address->prefixlen);
 }
 
 static void update_stats(struct connman_ipdevice *ipdevice,
 						struct rtnl_link_stats *stats)
 {
+	struct connman_service *service;
+
 	if (stats->rx_packets == 0 && stats->tx_packets == 0)
 		return;
 
@@ -429,7 +352,17 @@ static void update_stats(struct connman_ipdevice *ipdevice,
 	connman_info("%s {TX} %u packets %u bytes", ipdevice->ifname,
 					stats->tx_packets, stats->tx_bytes);
 
-	if (ipdevice->config == NULL)
+	if (ipdevice->config_ipv4 == NULL && ipdevice->config_ipv6 == NULL)
+		return;
+
+	if (ipdevice->config_ipv4)
+		service = connman_ipconfig_get_data(ipdevice->config_ipv4);
+	else if (ipdevice->config_ipv6)
+		service = connman_ipconfig_get_data(ipdevice->config_ipv6);
+	else
+		return;
+
+	if (service == NULL)
 		return;
 
 	ipdevice->rx_packets = stats->rx_packets;
@@ -441,7 +374,7 @@ static void update_stats(struct connman_ipdevice *ipdevice,
 	ipdevice->rx_dropped = stats->rx_dropped;
 	ipdevice->tx_dropped = stats->tx_dropped;
 
-	__connman_service_notify(ipdevice->config,
+	__connman_service_notify(service,
 				ipdevice->rx_packets, ipdevice->tx_packets,
 				ipdevice->rx_bytes, ipdevice->tx_bytes,
 				ipdevice->rx_errors, ipdevice->tx_errors,
@@ -592,6 +525,17 @@ void __connman_ipconfig_dellink(int index, struct rtnl_link_stats *stats)
 	g_hash_table_remove(ipdevice_hash, GINT_TO_POINTER(index));
 }
 
+static inline gint check_duplicate_address(gconstpointer a, gconstpointer b)
+{
+	const struct connman_ipaddress *addr1 = a;
+	const struct connman_ipaddress *addr2 = b;
+
+	if (addr1->prefixlen != addr2->prefixlen)
+		return addr2->prefixlen - addr1->prefixlen;
+
+	return g_strcmp0(addr1->local, addr2->local);
+}
+
 void __connman_ipconfig_newaddr(int index, int family, const char *label,
 				unsigned char prefixlen, const char *address)
 {
@@ -612,25 +556,29 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 	ipaddress->prefixlen = prefixlen;
 	ipaddress->local = g_strdup(address);
 
+	if (g_slist_find_custom(ipdevice->address_list, ipaddress,
+					check_duplicate_address)) {
+		connman_ipaddress_free(ipaddress);
+		return;
+	}
+
 	ipdevice->address_list = g_slist_append(ipdevice->address_list,
 								ipaddress);
 
-	connman_info("%s {add} address %s/%u label %s", ipdevice->ifname,
-						address, prefixlen, label);
+	connman_info("%s {add} address %s/%u label %s family %d",
+		ipdevice->ifname, address, prefixlen, label, family);
 
-	if (ipdevice->config != NULL) {
-		if (family == AF_INET6 && ipdevice->config->ipv6 != NULL)
-			connman_ipaddress_copy(ipdevice->config->ipv6->system,
-							ipaddress);
-		else
-			connman_ipaddress_copy(ipdevice->config->system,
-							ipaddress);
-	}
+	if (ipdevice->config_ipv4 != NULL && family == AF_INET)
+		connman_ipaddress_copy(ipdevice->config_ipv4->system,
+					ipaddress);
 
-	if ((ipdevice->flags & (IFF_RUNNING | IFF_LOWER_UP)) != (IFF_RUNNING | IFF_LOWER_UP))
+	else if (ipdevice->config_ipv6 != NULL && family == AF_INET6)
+		connman_ipaddress_copy(ipdevice->config_ipv6->system,
+					ipaddress);
+	else
 		return;
 
-	if (g_slist_length(ipdevice->address_list) > 1)
+	if ((ipdevice->flags & (IFF_RUNNING | IFF_LOWER_UP)) != (IFF_RUNNING | IFF_LOWER_UP))
 		return;
 
 	for (list = g_list_first(ipconfig_list); list;
@@ -707,19 +655,28 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 
 	if (scope == 0 && g_strcmp0(dst, "0.0.0.0") == 0) {
 		GSList *list;
+		GList *config_list;
 
 		if (family == AF_INET6) {
 			g_free(ipdevice->ipv6_gateway);
 			ipdevice->ipv6_gateway = g_strdup(gateway);
+
+			if (ipdevice->config_ipv6 != NULL &&
+				ipdevice->config_ipv6->system != NULL) {
+				g_free(ipdevice->config_ipv6->system->gateway);
+				ipdevice->config_ipv6->system->gateway =
+					g_strdup(gateway);
+			}
 		} else {
 			g_free(ipdevice->ipv4_gateway);
 			ipdevice->ipv4_gateway = g_strdup(gateway);
-		}
 
-		if (ipdevice->config != NULL &&
-					ipdevice->config->system != NULL) {
-			g_free(ipdevice->config->system->gateway);
-			ipdevice->config->system->gateway = g_strdup(gateway);
+			if (ipdevice->config_ipv4 != NULL &&
+				ipdevice->config_ipv4->system != NULL) {
+				g_free(ipdevice->config_ipv4->system->gateway);
+				ipdevice->config_ipv4->system->gateway =
+					g_strdup(gateway);
+			}
 		}
 
 		for (list = ipdevice->address_list; list; list = list->next) {
@@ -728,6 +685,20 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 			g_free(ipaddress->gateway);
 			ipaddress->gateway = g_strdup(gateway);
 		}
+
+		for (config_list = g_list_first(ipconfig_list); config_list;
+					config_list = g_list_next(config_list)) {
+			struct connman_ipconfig *ipconfig = config_list->data;
+
+			if (index != ipconfig->index)
+				continue;
+
+			if (ipconfig->ops == NULL)
+				continue;
+
+			if (ipconfig->ops->ip_bound)
+				ipconfig->ops->ip_bound(ipconfig);
+		}
 	}
 
 	connman_info("%s {add} route %s gw %s scope %u <%s>",
@@ -748,19 +719,26 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 
 	if (scope == 0 && g_strcmp0(dst, "0.0.0.0") == 0) {
 		GSList *list;
+		GList *config_list;
 
 		if (family == AF_INET6) {
 			g_free(ipdevice->ipv6_gateway);
 			ipdevice->ipv6_gateway = NULL;
+
+			if (ipdevice->config_ipv6 != NULL &&
+				ipdevice->config_ipv6->system != NULL) {
+				g_free(ipdevice->config_ipv6->system->gateway);
+				ipdevice->config_ipv6->system->gateway = NULL;
+			}
 		} else {
 			g_free(ipdevice->ipv4_gateway);
 			ipdevice->ipv4_gateway = NULL;
-		}
 
-		if (ipdevice->config != NULL &&
-					ipdevice->config->system != NULL) {
-			g_free(ipdevice->config->system->gateway);
-			ipdevice->config->system->gateway = NULL;
+			if (ipdevice->config_ipv4 != NULL &&
+				ipdevice->config_ipv4->system != NULL) {
+				g_free(ipdevice->config_ipv4->system->gateway);
+				ipdevice->config_ipv4->system->gateway = NULL;
+			}
 		}
 
 		for (list = ipdevice->address_list; list; list = list->next) {
@@ -769,6 +747,20 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 			g_free(ipaddress->gateway);
 			ipaddress->gateway = NULL;
 		}
+
+		for (config_list = g_list_first(ipconfig_list); config_list;
+					config_list = g_list_next(config_list)) {
+			struct connman_ipconfig *ipconfig = config_list->data;
+
+			if (index != ipconfig->index)
+				continue;
+
+			if (ipconfig->ops == NULL)
+				continue;
+
+			if (ipconfig->ops->ip_release)
+				ipconfig->ops->ip_release(ipconfig);
+		}
 	}
 
 	connman_info("%s {del} route %s gw %s scope %u <%s>",
@@ -827,9 +819,16 @@ const char *__connman_ipconfig_get_gateway(int index)
 	if (ipdevice->ipv4_gateway != NULL)
 		return ipdevice->ipv4_gateway;
 
-	if (ipdevice->config != NULL &&
-			ipdevice->config->address != NULL)
-		return ipdevice->config->address->gateway;
+	if (ipdevice->config_ipv4 != NULL &&
+			ipdevice->config_ipv4->address != NULL)
+		return ipdevice->config_ipv4->address->gateway;
+
+	if (ipdevice->ipv6_gateway != NULL)
+		return ipdevice->ipv6_gateway;
+
+	if (ipdevice->config_ipv6 != NULL &&
+			ipdevice->config_ipv6->address != NULL)
+		return ipdevice->config_ipv6->address->gateway;
 
 	return NULL;
 }
@@ -837,9 +836,6 @@ const char *__connman_ipconfig_get_gateway(int index)
 void __connman_ipconfig_set_index(struct connman_ipconfig *ipconfig, int index)
 {
 	ipconfig->index = index;
-
-	if (ipconfig->ipv6 != NULL)
-		ipconfig->ipv6->index = index;
 }
 
 static struct connman_ipconfig *create_ipv6config(int index)
@@ -852,6 +848,8 @@ static struct connman_ipconfig *create_ipv6config(int index)
 	if (ipv6config == NULL)
 		return NULL;
 
+	ipv6config->refcount = 1;
+
 	ipv6config->index = index;
 	ipv6config->type = CONNMAN_IPCONFIG_TYPE_IPV6;
 	ipv6config->method = CONNMAN_IPCONFIG_METHOD_OFF;
@@ -864,8 +862,6 @@ static struct connman_ipconfig *create_ipv6config(int index)
 
 	ipv6config->system = connman_ipaddress_alloc(AF_INET6);
 
-	ipv6config->ipv6 = NULL;
-
 	DBG("ipconfig %p", ipv6config);
 
 	return ipv6config;
@@ -878,10 +874,14 @@ static struct connman_ipconfig *create_ipv6config(int index)
  *
  * Returns: a newly-allocated #connman_ipconfig structure
  */
-struct connman_ipconfig *connman_ipconfig_create(int index)
+struct connman_ipconfig *connman_ipconfig_create(int index,
+					enum connman_ipconfig_type type)
 {
 	struct connman_ipconfig *ipconfig;
 
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
+		return create_ipv6config(index);
+
 	DBG("index %d", index);
 
 	ipconfig = g_try_new0(struct connman_ipconfig, 1);
@@ -901,38 +901,11 @@ struct connman_ipconfig *connman_ipconfig_create(int index)
 
 	ipconfig->system = connman_ipaddress_alloc(AF_INET);
 
-	ipconfig->ipv6 = create_ipv6config(index);
-
 	DBG("ipconfig %p", ipconfig);
 
 	return ipconfig;
 }
 
-/**
- * connman_ipconfig_clone:
- *
- * Clone an ipconfig structure and create new reference.
- *
- * Returns: a newly-allocated #connman_ipconfig structure
- */
-struct connman_ipconfig *connman_ipconfig_clone(struct connman_ipconfig *ipconfig)
-{
-	struct connman_ipconfig *ipconfig_clone;
-
-	DBG("ipconfig %p", ipconfig);
-
-	ipconfig_clone = g_try_new0(struct connman_ipconfig, 1);
-	if (ipconfig_clone == NULL)
-		return NULL;
-
-	ipconfig_clone->refcount = 1;
-
-	ipconfig_clone->origin = connman_ipconfig_ref(ipconfig);
-
-	ipconfig_clone->index = -1;
-
-	return ipconfig_clone;
-}
 
 /**
  * connman_ipconfig_ref:
@@ -947,17 +920,6 @@ struct connman_ipconfig *connman_ipconfig_ref(struct connman_ipconfig *ipconfig)
 	return ipconfig;
 }
 
-static void  free_ipv6config(struct connman_ipconfig *ipconfig)
-{
-	if (ipconfig == NULL)
-		return;
-
-	connman_ipconfig_set_ops(ipconfig, NULL);
-	connman_ipaddress_free(ipconfig->system);
-	connman_ipaddress_free(ipconfig->address);
-	g_free(ipconfig->ipv6);
-}
-
 /**
  * connman_ipconfig_unref:
  * @ipconfig: ipconfig structure
@@ -966,7 +928,8 @@ static void  free_ipv6config(struct connman_ipconfig *ipconfig)
  */
 void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
 {
-	if (g_atomic_int_dec_and_test(&ipconfig->refcount) == TRUE) {
+	if (ipconfig &&
+		g_atomic_int_dec_and_test(&ipconfig->refcount) == TRUE) {
 		__connman_ipconfig_disable(ipconfig);
 
 		connman_ipconfig_set_ops(ipconfig, NULL);
@@ -978,7 +941,6 @@ void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
 
 		connman_ipaddress_free(ipconfig->system);
 		connman_ipaddress_free(ipconfig->address);
-		free_ipv6config(ipconfig->ipv6);
 		g_free(ipconfig);
 	}
 }
@@ -991,6 +953,9 @@ void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
  */
 void *connman_ipconfig_get_data(struct connman_ipconfig *ipconfig)
 {
+	if (ipconfig == NULL)
+		return NULL;
+
 	return ipconfig->ops_data;
 }
 
@@ -1060,15 +1025,6 @@ void connman_ipconfig_set_ops(struct connman_ipconfig *ipconfig,
 	ipconfig->ops = ops;
 }
 
-struct connman_ipconfig *connman_ipconfig_get_ipv6config(
-				struct connman_ipconfig *ipconfig)
-{
-	if (ipconfig == NULL)
-		return NULL;
-
-	return ipconfig->ipv6;
-}
-
 /**
  * connman_ipconfig_set_method:
  * @ipconfig: ipconfig structure
@@ -1115,7 +1071,8 @@ void __connman_ipconfig_set_element_ipv6_gateway(
 			struct connman_ipconfig *ipconfig,
 				struct connman_element *element)
 {
-	element->ipv6.gateway = ipconfig->ipv6->address->gateway;
+	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6)
+		element->ipv6.gateway = ipconfig->address->gateway;
 }
 
 /*
@@ -1133,8 +1090,11 @@ int __connman_ipconfig_set_gateway(struct connman_ipconfig *ipconfig,
 
 	connection->type  = CONNMAN_ELEMENT_TYPE_CONNECTION;
 	connection->index = ipconfig->index;
-	connection->ipv4.gateway = ipconfig->address->gateway;
-	connection->ipv6.gateway = ipconfig->ipv6->address->gateway;
+
+	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		connection->ipv4.gateway = ipconfig->address->gateway;
+	else if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6)
+		connection->ipv6.gateway = ipconfig->address->gateway;
 
 	if (connman_element_register(connection, parent) < 0)
 		connman_element_unref(connection);
@@ -1181,7 +1141,8 @@ int __connman_ipconfig_clear_address(struct connman_ipconfig *ipconfig)
 		break;
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
 		if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4)
-			return connman_inet_clear_address(ipconfig->index);
+			return connman_inet_clear_address(ipconfig->index,
+							ipconfig->address);
 		else if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6)
 			return connman_inet_clear_ipv6_address(
 						ipconfig->index,
@@ -1235,6 +1196,7 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 	struct connman_ipdevice *ipdevice;
 	gboolean up = FALSE, down = FALSE;
 	gboolean lower_up = FALSE, lower_down = FALSE;
+	enum connman_ipconfig_type type;
 
 	DBG("ipconfig %p", ipconfig);
 
@@ -1246,18 +1208,41 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 	if (ipdevice == NULL)
 		return -ENXIO;
 
-	if (ipdevice->config == ipconfig)
-		return -EALREADY;
+	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4) {
+		if (ipdevice->config_ipv4 == ipconfig)
+			return -EALREADY;
+		type = CONNMAN_IPCONFIG_TYPE_IPV4;
+	} else if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6) {
+		if (ipdevice->config_ipv6 == ipconfig)
+			return -EALREADY;
+		type = CONNMAN_IPCONFIG_TYPE_IPV6;
+	} else
+		return -EINVAL;
 
-	if (ipdevice->config != NULL) {
-		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV4 &&
+					ipdevice->config_ipv4 != NULL) {
+		ipconfig_list = g_list_remove(ipconfig_list,
+							ipdevice->config_ipv4);
 
-		connman_ipaddress_clear(ipdevice->config->system);
+		connman_ipaddress_clear(ipdevice->config_ipv4->system);
 
-		connman_ipconfig_unref(ipdevice->config);
+		connman_ipconfig_unref(ipdevice->config_ipv4);
 	}
 
-	ipdevice->config = connman_ipconfig_ref(ipconfig);
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV6 &&
+					ipdevice->config_ipv6 != NULL) {
+		ipconfig_list = g_list_remove(ipconfig_list,
+							ipdevice->config_ipv6);
+
+		connman_ipaddress_clear(ipdevice->config_ipv6->system);
+
+		connman_ipconfig_unref(ipdevice->config_ipv6);
+	}
+
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		ipdevice->config_ipv4 = connman_ipconfig_ref(ipconfig);
+	else if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
+		ipdevice->config_ipv6 = connman_ipconfig_ref(ipconfig);
 
 	ipconfig_list = g_list_append(ipconfig_list, ipconfig);
 
@@ -1299,18 +1284,28 @@ int __connman_ipconfig_disable(struct connman_ipconfig *ipconfig)
 	if (ipdevice == NULL)
 		return -ENXIO;
 
-	if (ipdevice->config == NULL || ipdevice->config != ipconfig)
+	if (ipdevice->config_ipv4 == NULL && ipdevice->config_ipv6 == NULL)
 		return -EINVAL;
 
-	ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
+	if (ipdevice->config_ipv4 == ipconfig) {
+		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
 
-	connman_ipaddress_clear(ipdevice->config->system);
-	connman_ipaddress_clear(ipdevice->config->ipv6->system);
+		connman_ipaddress_clear(ipdevice->config_ipv4->system);
+		connman_ipconfig_unref(ipdevice->config_ipv4);
+		ipdevice->config_ipv4 = NULL;
+		return 0;
+	}
 
-	connman_ipconfig_unref(ipdevice->config);
-	ipdevice->config = NULL;
+	if (ipdevice->config_ipv6 == ipconfig) {
+		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
 
-	return 0;
+		connman_ipaddress_clear(ipdevice->config_ipv6->system);
+		connman_ipconfig_unref(ipdevice->config_ipv6);
+		ipdevice->config_ipv6 = NULL;
+		return 0;
+	}
+
+	return -EINVAL;
 }
 
 const char *__connman_ipconfig_method2string(enum connman_ipconfig_method method)
@@ -1352,6 +1347,9 @@ void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
 
 	DBG("");
 
+	if (ipconfig->type != CONNMAN_IPCONFIG_TYPE_IPV4)
+		return;
+
 	str = __connman_ipconfig_method2string(ipconfig->method);
 	if (str == NULL)
 		return;
@@ -1388,6 +1386,9 @@ void __connman_ipconfig_append_ipv6(struct connman_ipconfig *ipconfig,
 
 	DBG("");
 
+	if (ipconfig->type != CONNMAN_IPCONFIG_TYPE_IPV6)
+		return;
+
 	str = __connman_ipconfig_method2string(ipconfig->method);
 	if (str == NULL)
 		return;
@@ -1603,112 +1604,6 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 	return 0;
 }
 
-void __connman_ipconfig_append_proxy(struct connman_ipconfig *ipconfig,
-							DBusMessageIter *iter)
-{
-	struct connman_ipdevice *ipdevice;
-	const char *method = "direct";
-
-	ipdevice = g_hash_table_lookup(ipdevice_hash,
-                                        GINT_TO_POINTER(ipconfig->index));
-	if (ipdevice == NULL)
-		goto done;
-
-	if (ipdevice->pac == NULL)
-		goto done;
-
-	method = "auto-config";
-
-	connman_dbus_dict_append_basic(iter, "URL",
-					DBUS_TYPE_STRING, &ipdevice->pac);
-
-done:
-	connman_dbus_dict_append_basic(iter, "Method",
-						DBUS_TYPE_STRING, &method);
-}
-
-void __connman_ipconfig_append_proxyconfig(struct connman_ipconfig *ipconfig,
-							DBusMessageIter *iter)
-{
-	struct connman_ipdevice *ipdevice;
-	const char *method = "auto";
-
-	ipdevice = g_hash_table_lookup(ipdevice_hash,
-					GINT_TO_POINTER(ipconfig->index));
-	if (ipdevice == NULL)
-		goto done;
-
-	if (ipdevice->proxy == NULL)
-		goto done;
-
-	method = ipdevice->proxy;
-
-done:
-	connman_dbus_dict_append_basic(iter, "Method",
-						DBUS_TYPE_STRING, &method);
-}
-
-int __connman_ipconfig_set_proxyconfig(struct connman_ipconfig *ipconfig,
-							DBusMessageIter *array)
-{
-	struct connman_ipdevice *ipdevice;
-	DBusMessageIter dict;
-	const char *method;
-
-	DBG("ipconfig %p", ipconfig);
-
-	ipdevice = g_hash_table_lookup(ipdevice_hash,
-					GINT_TO_POINTER(ipconfig->index));
-	if (ipdevice == NULL)
-		return -ENXIO;
-
-	if (dbus_message_iter_get_arg_type(array) != DBUS_TYPE_ARRAY)
-		return -EINVAL;
-
-	dbus_message_iter_recurse(array, &dict);
-
-	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
-		DBusMessageIter entry;
-		const char *key;
-		int type;
-
-		dbus_message_iter_recurse(&dict, &entry);
-
-		if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_STRING)
-			return -EINVAL;
-
-		dbus_message_iter_get_basic(&entry, &key);
-		dbus_message_iter_next(&entry);
-
-		type = dbus_message_iter_get_arg_type(&entry);
-
-		if (g_str_equal(key, "Method") == TRUE) {
-			if (type != DBUS_TYPE_STRING)
-				return -EINVAL;
-
-			dbus_message_iter_get_basic(&entry, &method);
-			if (strlen(method) == 0)
-				method = NULL;
-		}
-
-		dbus_message_iter_next(&dict);
-	}
-
-	DBG("method %s", method);
-
-	if (method == NULL)
-		return -EINVAL;
-
-	if (g_str_equal(method, "auto") == FALSE &&
-				g_str_equal(method, "direct") == FALSE)
-		return -EINVAL;
-
-	g_free(ipdevice->proxy);
-	ipdevice->proxy = g_strdup(method);
-
-	return 0;
-}
-
 void __connman_ipconfig_append_ethernet(struct connman_ipconfig *ipconfig,
 							DBusMessageIter *iter)
 {
@@ -1753,6 +1648,10 @@ int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 			ipconfig->method = CONNMAN_IPCONFIG_METHOD_OFF;
 	} else
 		ipconfig->method = __connman_ipconfig_string2method(method);
+
+	if (ipconfig->method == CONNMAN_IPCONFIG_METHOD_UNKNOWN)
+		ipconfig->method = CONNMAN_IPCONFIG_METHOD_OFF;
+
 	g_free(key);
 
 	key = g_strdup_printf("%snetmask_prefixlen", prefix);
